#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>

#include "qemu/osdep.h"
#include "guest-agent-core.h"

#include "aux-io.h"

#define BUFF_READ_SIZE  (64 * 1024)
#define BUFF_HIGH_MARK  (128 * 1024)
#define BUFF_LOW_MARK   (16 * 1024)

enum {
    AUX_HEADER,
    AUX_DATA
};

typedef struct AuxIoBuf {
    char *buffer;
    size_t size;
    size_t length;
    bool eof;
    int cid;
} AuxIoBuf;

typedef struct AuxOutputWatcher {
    guint source;
    GList *queue;
} AuxOutputWatcher;

static void output_watcher_free(void *opaque)
{
    g_slice_free(AuxOutputWatcher, opaque);
}

static GHashTable *output_watchers; /* (GIOChannel *) -> (AuxOutputWatcher *) */
static GHashTable *ga_aux_channels;

static void aux_input_cb(const char *buffer, size_t size, void *cbdata);
static gboolean aux_output_watch(GIOChannel *chan, GIOCondition cond, gpointer d);
static gboolean aux_input_watch(GIOChannel *chan, GIOCondition cond, gpointer d);
static void aux_slice(AuxIoChannel *ch, char *buffer, size_t size, bool eof);

void aux_init(void)
{
    ga_aux_channels = g_hash_table_new(g_direct_hash, g_direct_equal);
    output_watchers = g_hash_table_new_full(g_direct_hash, g_direct_equal,
                                            NULL, output_watcher_free);
}

static gboolean aux_input_watch(GIOChannel *chan, GIOCondition cond, gpointer d)
{
    AuxIoChannel *ch = g_hash_table_lookup(ga_aux_channels, d);
    if (!ch) {
        return false;
    }

    gsize bytes_read;
    GIOStatus gstatus;
    bool ret = false;

    AuxMessage *msg = (AuxMessage *)g_new(char, BUFF_READ_SIZE + sizeof(msg->header));
    msg->header.magic = AUX_MAGIC_NUMBER;
    msg->header.length = 0;
    msg->header.cid = ch->cid;

    if (cond & G_IO_IN) {
        gstatus = g_io_channel_read_chars(chan, msg->data, BUFF_READ_SIZE,
                &bytes_read, NULL);
        if (gstatus == G_IO_STATUS_NORMAL) {
            msg->header.length = bytes_read;
            ch->total += bytes_read;
            ret = true;
        } else {
            ch->revent = 0;
        }
    } else if (cond & (G_IO_HUP|G_IO_ERR)) {
        if (cond & G_IO_ERR) {
            g_debug("aux i/o err %s cond=%x", ch->name, cond);
        }
        ch->revent = 0;
    }

    if (ch->wchan && ch->size > BUFF_HIGH_MARK) {
        g_debug("aux flow stop reader %s queue=%u", ch->name, (int)ch->size);
        ch->revent = 0;
        ret = false;
    }

#ifdef DEBUG_AUX_IO
    static uint32_t seq;
    msg->header.seq = seq++;
    g_debug("aux header %s: len=%u, seq=%u",
            ch->name, msg->header.length, msg->header.seq);
#endif

    aux_slice(ch, (void*)msg,
        sizeof(msg->header) + msg->header.length,  msg->header.length == 0);

    return ret;
}

static void output_queue_pop(AuxOutputWatcher *watcher)
{
    if (watcher->queue == NULL) {
        return;
    }

    AuxIoBuf *p = watcher->queue->data;

    watcher->queue = g_list_delete_link(watcher->queue, watcher->queue);

    g_free(p->buffer);
    g_slice_free(AuxIoBuf, p);
}

static gboolean aux_output_watch(GIOChannel *chan, GIOCondition cond, gpointer d)
{
    (void) d;
    AuxOutputWatcher *watcher = g_hash_table_lookup(output_watchers, chan);

    g_assert(watcher != NULL);

    if (watcher->queue == NULL) {
        g_hash_table_remove(output_watchers, chan);
        return false;
    }

    AuxIoBuf *p = watcher->queue->data;
    AuxIoChannel *ch = g_hash_table_lookup(ga_aux_channels,
                                           GINT_TO_POINTER(p->cid));

    if (!ch) {
        output_queue_pop(watcher);
        return true;
    }

    if (cond & G_IO_ERR) {
        g_debug("aux i/o err %s cond=%x", ch->name, cond);
        aux_remove_channel(ch->cid);
        output_queue_pop(watcher);
        return true;
    }

    GIOStatus ret;
    GError * gerr = NULL;
    gsize bytes_written = 0;

    ret = g_io_channel_write_chars(ch->wchan, (gchar *)p->buffer + p->length,
                p->size - p->length, &bytes_written, &gerr);
    if (ret != G_IO_STATUS_NORMAL) {
        return true; /* Try again later */
    }

    g_debug("aux async_write %s bytes=%d written=%d offset=%d queue=%d ret=%d",
        ch->name, (int)(p->size - p->length), (int)bytes_written, (int)p->length, (int)ch->size, ret);

    ch->size -= bytes_written;
    p->length += bytes_written;

    if (p->size == p->length)  {
        bool eof = p->eof;

        output_queue_pop(watcher);
        if (eof) {
            g_debug("aux eof %s queue=%u", ch->name, (int)ch->size);
            aux_remove_channel(ch->cid);
            return true;
        }
    }

    /* if have free space, resume readers */
    if (ch->size <= BUFF_LOW_MARK) {
        if (ch->rchan) {
            /* this is guest out/err, add watch if needed */
            if (!ch->revent) {
                g_debug("aux start reader %s queue=%u", ch->name, (int)ch->size);
                ch->revent = g_io_add_watch(ch->rchan, G_IO_IN|G_IO_HUP|G_IO_ERR, aux_input_watch,
                        GINT_TO_POINTER(ch->cid));
            }
        } else {
            /* this is stdin, resume aux listening */
            if (ch->overflow) {
                g_debug("aux flow start %s queue=%u", ch->name, (int)ch->size);
                ch->overflow = false;
                ga_channel_flow_control(ga_get_aux_channel(ga_state), true);
            }
        }
    }

    return true;
}

static void aux_slice(AuxIoChannel *ch, char *buffer, size_t size, bool eof)
{
    if (ch->wchan) {
        /* async write to glib channel */

        AuxIoBuf *p = g_slice_new(AuxIoBuf);

        p->buffer = (char *)buffer;
        p->size = size;
        p->length = 0;
        p->eof = eof;
        p->cid = ch->cid;

        /* statistics */
        ch->size += size;

        g_debug("aux slice %s size=%u queue=%u eof=%d", ch->name, (int)size, (int)ch->size, eof);

        AuxOutputWatcher *watcher = g_hash_table_lookup(output_watchers,
                                                        ch->wchan);
        if (!watcher) {
            watcher = g_slice_new0(AuxOutputWatcher);
            watcher->source = g_io_add_watch(ch->wchan, G_IO_OUT | G_IO_ERR,
                                             aux_output_watch, NULL);
            g_hash_table_insert(output_watchers, ch->wchan, watcher);
        }
        watcher->queue = g_list_append(watcher->queue, p);

        if (!ch->rchan) {
            /* for stdin listener, stop aux */
            if (!ch->overflow && ch->size > BUFF_HIGH_MARK) {
                g_debug("aux flow stop %s queue=%u", ch->name, (int)ch->size);
                ch->overflow = true;
                ga_channel_flow_control(ga_get_aux_channel(ga_state), false);
            }
        }
    } else {
        /* sync write to GA channel - async not implemented */

        g_debug("aux sync_write %s size=%u", ch->name, (int)size);
        GIOStatus status = ga_channel_write_all(ga_get_aux_channel(ga_state), buffer, size);
        g_debug("aux sync_write %s status=%u", ch->name, status);

        if (eof) {
            g_debug("aux eof %s", ch->name);
            aux_remove_channel(ch->cid);
        }
        if (status != G_IO_STATUS_NORMAL) {
            slog("I/O error writing to aux channel");
        }

        g_free(buffer);
    }
}

static void aux_dispatch_header(AuxMessageParser *parser, const char *buffer, size_t size)
{
    do {
        size_t len = sizeof(parser->header) - parser->header_len;
        if (size < len) {
            len = size;
        }

        memcpy((char*)&parser->header + parser->header_len, buffer, len);
        buffer += len;
        size -= len;

        parser->header_len += len;

        if (parser->header_len >= sizeof(parser->header.magic)) {
            if (parser->header.magic == AUX_MAGIC_NUMBER) {
                /* valid magic, try to use this header */
                break;
            } else {
                /* invalid magic, skip one byte and retry */
                parser->header_len--;
                memmove(&parser->header, (char*)&parser->header + 1, parser->header_len);
            }
        }
    } while (size != 0);

    if (sizeof(parser->header) == parser->header_len) {
        g_debug("aux dispatch_header cid=%u len=%u", parser->header.cid, parser->header.length);

        if (parser->header.length == 0) {
            /* got EOF packet, no more data coming - process it now */

            AuxIoChannel *ch = g_hash_table_lookup(ga_aux_channels, GINT_TO_POINTER(parser->header.cid));

            g_debug("aux header cid=%u EOF %s", parser->header.cid, ch ? "" : "spurious");

            if (ch) {
                aux_slice(ch, NULL, 0, true);
            }

            /* packet processed, so expect next header */
            parser->state = AUX_HEADER;
            memset(&parser->header, 0, sizeof(parser->header));
            parser->header_len = 0;
        } else {
            /* got complete header, switch state to data */
            parser->state = AUX_DATA;
            parser->data_len = 0;
        }
    }

    if (size != 0) {
        /* parse the rest as data */
        aux_message_parser_feed(parser, buffer, size);
    }
}

static void aux_dispatch_data(AuxMessageParser *parser, const char *buffer, size_t size)
{
    size_t len;

    /* channel may be closed, so we check it here for each data chunk */
    AuxIoChannel *ch = g_hash_table_lookup(ga_aux_channels, GINT_TO_POINTER(parser->header.cid));

    len = MIN(parser->header.length - parser->data_len, size);
    if (ch) {
        ch->total += len;
        ch->cb(buffer, len, ch);
    } else {
        g_debug("aux dispatch_data cid=%u len=%u size=%u spurious",
            parser->header.cid, parser->header.length, (int)size);
    }

    buffer += len;
    size -= len;
    parser->data_len += len;

    if (parser->data_len == parser->header.length) {
        /* if processed complete data, switch state to header */
        parser->state = AUX_HEADER;
        memset(&parser->header, 0, sizeof(parser->header));
        parser->header_len = 0;
    }

    if (size != 0) {
        /* parse the rest as header */
        aux_message_parser_feed(parser, buffer, size);
    }
}

void aux_message_parser_feed(AuxMessageParser *parser, const char *buffer, size_t size)
{
    switch (parser->state)
    {
        case AUX_HEADER:
            aux_dispatch_header(parser, buffer, size);
            break;

        case AUX_DATA:
            aux_dispatch_data(parser, buffer, size);
            break;
    }
}

static void aux_input_cb(const char *buffer, size_t size, void *cbdata)
{
    aux_slice(cbdata, g_memdup(buffer, size), size, false);
}

void aux_add_channel(int fd, guint cid, GIOChannel **ch)
{
    const char *fdstr[] = { "stdin", "stdout", "stderr" };
    AuxIoChannel *p = g_new0(AuxIoChannel, 1);

    sprintf(p->name, "[%s cid=%u]", fdstr[fd], cid);

    g_debug("aux add_channel %s", p->name);

    p->cid = cid;
    p->pchan = ch;
    if (fd == 0) {
        /* stdin */
        p->wchan = *ch;
        p->cb = aux_input_cb;
    } else {
        /* stdout/stderr */
        p->rchan = *ch;
#ifndef G_OS_WIN32
        p->wchan = ga_channel_get_io_channel(ga_get_aux_channel(ga_state));
#endif
        p->revent = g_io_add_watch(*ch, G_IO_IN|G_IO_HUP|G_IO_ERR, aux_input_watch,
            GINT_TO_POINTER(cid));
    }

    g_hash_table_insert(ga_aux_channels, GINT_TO_POINTER(cid), p);
}

void aux_remove_channel(guint cid)
{
    AuxIoChannel *p = g_hash_table_lookup(ga_aux_channels, GINT_TO_POINTER(cid));
    if (p) {
        slog("guest-exec close %s queue=%d total=%d", p->name, (int)p->size, (int)p->total);

        if (p->overflow) {
            g_debug("aux flow start %s on close", p->name);
            ga_channel_flow_control(ga_get_aux_channel(ga_state), true);
        }

        g_hash_table_remove(ga_aux_channels, GINT_TO_POINTER(cid));

        if (p->revent) {
            g_source_remove(p->revent);
        }

	if (p->pchan && g_atomic_pointer_get(p->pchan)) {
            g_io_channel_shutdown(*p->pchan, true, NULL);
            g_io_channel_unref(*p->pchan);
            g_atomic_pointer_set(p->pchan, NULL);
	}

        g_free(p);
    }
}
