/*
 * QEMU Guest Agent common/cross-platform command implementations
 *
 * Copyright IBM Corp. 2012
 *
 * Authors:
 *  Michael Roth      <mdroth@linux.vnet.ibm.com>
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */

#include "qemu/osdep.h"
#include "qemu/units.h"
#include "qga/guest-agent-core.h"
#include "qga-qapi-commands.h"
#include "qapi/error.h"
#include "qapi/qmp/qerror.h"
#include "qemu/base64.h"
#include "qemu/cutils.h"
#include "qemu/atomic.h"
#include "commands-common.h"
#include "aux-io.h"

#include <stdio.h>
#ifndef G_OS_WIN32
#ifdef __FreeBSD__
#include <sys/types.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <libutil.h>
#else
#include <pty.h>
#endif
#endif

/* Maximum captured guest-exec out_data/err_data - 16MB */
#define GUEST_EXEC_MAX_OUTPUT (16 * 1024 * 1024)
/* Allocation and I/O buffer for reading guest-exec out_data/err_data - 4KB */
#define GUEST_EXEC_IO_SIZE (4 * 1024)
/*
 * Maximum file size to read - 48MB
 *
 * (48MB + Base64 3:4 overhead = JSON parser 64 MB limit)
 */
#define GUEST_FILE_READ_COUNT_MAX (48 * MiB)

#ifdef G_OS_WIN32
#  define SHELL_CMD "cmd"
#  define SHELL_ARG "/c"
#else
#  define SHELL_CMD "bash"
#  define SHELL_ARG "-c"
#endif

/* Note: in some situations, like with the fsfreeze, logging may be
 * temporarilly disabled. if it is necessary that a command be able
 * to log for accounting purposes, check ga_logging_enabled() beforehand,
 * and use the QERR_QGA_LOGGING_DISABLED to generate an error
 */
void slog(const gchar *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    g_logv("syslog", G_LOG_LEVEL_INFO, fmt, ap);
    va_end(ap);
}

int64_t qmp_guest_sync_delimited(int64_t id, Error **errp)
{
    ga_set_response_delimited(ga_state);
    return id;
}

int64_t qmp_guest_sync(int64_t id, Error **errp)
{
    return id;
}

void qmp_guest_ping(Error **errp)
{
    slog("guest-ping called");
}

static void qmp_command_info(const QmpCommand *cmd, void *opaque)
{
    GuestAgentInfo *info = opaque;
    GuestAgentCommandInfo *cmd_info;

    cmd_info = g_new0(GuestAgentCommandInfo, 1);
    cmd_info->name = g_strdup(qmp_command_name(cmd));
    cmd_info->enabled = qmp_command_is_enabled(cmd);
    cmd_info->success_response = qmp_has_success_response(cmd);

    QAPI_LIST_PREPEND(info->supported_commands, cmd_info);
}

/*
 * Try to read tools version from VERSION file once and cache it for future
 * usage.  Return strdup'ed string containing the version, so the caller
 * is responsible for freeing it.
 */
char *get_guesttools_version(void)
{
    enum {version_buf_size = 128};
    static char version[version_buf_size]; /* initialized to NULL by default */

    FILE *version_file;
    size_t version_len;
#ifdef G_OS_WIN32
    const char *version_file_path = "C:\\Program Files\\Qemu-ga\\VERSION";
#else
    const char *version_file_path = "/usr/share/qemu-ga/VERSION";
#endif

    if (version[0]) {
        return g_strdup(version);
    }

    /* First try to read version from the VERSION file */
    version_file = fopen(version_file_path, "r");
    if (version_file) {
        if (fgets(version, version_buf_size, version_file)) {
            version_len = strlen(version);
            if (version[version_len - 1] == '\n') {
                version[version_len - 1] = '\0';
            }
        }
        fclose(version_file);
    } else {
        slog("qga: unable to open file %s: %s", version_file_path,
             g_strerror(errno));
    }

    /* In case of failure, fallback to internal QEMU version report */
    if (!version[0]) {
        g_assert(strlen(QEMU_VERSION) < version_buf_size);
        strcpy(version, QEMU_VERSION);
    }

    return g_strdup(version);
}

struct GuestAgentInfo *qmp_guest_info(Error **errp)
{
    GuestAgentInfo *info = g_new0(GuestAgentInfo, 1);
    info->version = get_guesttools_version();

    qmp_for_each_command(&ga_commands, qmp_command_info, info);
    return info;
}

struct GuestExecIOData {
    guchar *data;
    gsize size;
    gsize length;
    bool truncated;
    const char *name;
    GIOChannel *ch;
    guint cid;
};
typedef struct GuestExecIOData GuestExecIOData;

struct GuestExecInfo {
    GPid pid;
    int64_t pid_numeric;
    gint status;
    bool has_output;
    bool has_errput;
    bool with_pty;
    int  master_pty;
    int  slave_pty;
    bool finished;
    GuestExecIOData in;
    GuestExecIOData out;
    GuestExecIOData err;
    QTAILQ_ENTRY(GuestExecInfo) next;
};
typedef struct GuestExecInfo GuestExecInfo;

static struct {
    QTAILQ_HEAD(, GuestExecInfo) processes;
} guest_exec_state = {
    .processes = QTAILQ_HEAD_INITIALIZER(guest_exec_state.processes),
};

static int64_t gpid_to_int64(GPid pid)
{
#ifdef G_OS_WIN32
    return GetProcessId(pid);
#else
    return (int64_t)pid;
#endif
}

static void guest_exec_info_add(GuestExecInfo *gei, GPid pid)
{
    gei->pid = pid;
    gei->pid_numeric = gpid_to_int64(pid);
    QTAILQ_INSERT_TAIL(&guest_exec_state.processes, gei, next);
}

static GuestExecInfo *guest_exec_info_find(int64_t pid, Error **err)
{
    GuestExecInfo *gei;

    QTAILQ_FOREACH(gei, &guest_exec_state.processes, next) {
        if (gei->pid_numeric == pid) {
            return gei;
        }
    }

    error_setg(err, "Guest agent command failed, pid: %u not found",
        (uint32_t)pid);

    return NULL;
}

GuestExecStatus *qmp_guest_exec_status(int64_t pid, Error **errp)
{
    GuestExecInfo *gei;
    GuestExecStatus *ges;

    gei = guest_exec_info_find(pid, errp);
    if (gei == NULL) {
        slog("guest-exec-status failed, pid: %u not found", (uint32_t)pid);
        return NULL;
    }

    ges = g_new0(GuestExecStatus, 1);

    ges->exited = qatomic_mb_read(&gei->finished);
    bool out_fin = !gei->has_output || g_atomic_pointer_get(&gei->out.ch) == NULL;
    bool err_fin = !gei->has_errput || g_atomic_pointer_get(&gei->err.ch) == NULL;

    slog("guest-exec-status called, pid: %u fin: %d out: %d err: %d", (uint32_t)pid,
        (int)ges->exited, (int)out_fin, (int) err_fin);

    ges->exited &= out_fin && err_fin;

    if (ges->exited) {
        /* Glib has no portable way to parse exit status.
         * On UNIX, we can get either exit code from normal termination
         * or signal number.
         * On Windows, it is either the same exit code or the exception
         * value for an unhandled exception that caused the process
         * to terminate.
         * See MSDN for GetExitCodeProcess() and ntstatus.h for possible
         * well-known codes, e.g. C0000005 ACCESS_DENIED - analog of SIGSEGV
         * References:
         *   https://msdn.microsoft.com/en-us/library/windows/desktop/ms683189(v=vs.85).aspx
         *   https://msdn.microsoft.com/en-us/library/aa260331(v=vs.60).aspx
         */
#ifdef G_OS_WIN32
        /* Additionally WIN32 does not provide any additional information
         * on whether the child exited or terminated via signal.
         * We use this simple range check to distinguish application exit code
         * (usually value less then 256) and unhandled exception code with
         * ntstatus (always value greater then 0xC0000005). */
        if ((uint32_t)gei->status < 0xC0000000U) {
            ges->has_exitcode = true;
            ges->exitcode = gei->status;
        } else {
            ges->has_signal = true;
            ges->signal = gei->status;
        }
#else
        if (WIFEXITED(gei->status)) {
            ges->has_exitcode = true;
            ges->exitcode = WEXITSTATUS(gei->status);
        } else if (WIFSIGNALED(gei->status)) {
            ges->has_signal = true;
            ges->signal = WTERMSIG(gei->status);
        }
#endif
        if (gei->out.length > 0) {
            ges->has_out_data = true;
            ges->out_data = g_base64_encode(gei->out.data, gei->out.length);
            g_free(gei->out.data);
            ges->has_out_truncated = gei->out.truncated;
        }

        if (gei->err.length > 0) {
            ges->has_err_data = true;
            ges->err_data = g_base64_encode(gei->err.data, gei->err.length);
            g_free(gei->err.data);
            ges->has_err_truncated = gei->err.truncated;
        }

        slog("guest-exec-status finished, pid: %u, exitcode=%u", (uint32_t)pid, (uint32_t)ges->exitcode);

        QTAILQ_REMOVE(&guest_exec_state.processes, gei, next);

        g_free(gei);
    }

    return ges;
}

/* Get environment variables or arguments array for execve(). */
static char **guest_exec_get_args(const strList *entry, bool log)
{
    const strList *it;
    int count = 1, i = 0;  /* reserve for NULL terminator */
    char **args;
    char *str; /* for logging array of arguments */
    size_t str_size = 1;

    for (it = entry; it != NULL; it = it->next) {
        count++;
        str_size += 1 + strlen(it->value);
    }

    str = g_malloc(str_size);
    *str = 0;
    args = g_new(char *, count);
    for (it = entry; it != NULL; it = it->next) {
        args[i++] = it->value;
        pstrcat(str, str_size, it->value);
        if (it->next) {
            pstrcat(str, str_size, " ");
        }
    }
    args[i] = NULL;

    if (log) {
        slog("guest-exec called: \"%s\"", str);
    }
    g_free(str);

    return args;
}

#ifndef G_OS_WIN32
/*
 * Whitespace-only quooting for passing args to bash -c command-line.
 *
 * We do not quote a single space in order to maintain args separation.
 */
static gchar *my_shell_quote (const gchar *unquoted_string)
{
    const gchar *p;
    GString *dest;

    g_return_val_if_fail (unquoted_string != NULL, NULL);

    dest = g_string_new(NULL);

    p = unquoted_string;

    while (*p) {
        if (*p == ' ') {
            g_string_append(dest, "\\ ");
        } else {
            g_string_append_c(dest, *p);
        }
        ++p;
    }

    return g_string_free(dest, FALSE);
}

/*
 * concat_args - convert argv style list to a single string for
 * execution with "bash -c command-line" syntax.
 *
 * returns string value that must be g_free()'ed
 */
static char * concat_args(const strList * args)
{
    const strList *it;
    size_t size = 0;

    /*
     * If we have only one arg, it is already a string
     * ready to be interpreted with bash -c
     */
    if (args->next == NULL) {
        return g_strdup(args->value);
    }

    for (it = args; it != NULL; it = it->next) {
        char * t = my_shell_quote(it->value);
        size += strlen(t) + 1;
        g_free(t);
    }

    gchar * s = g_malloc(size);
    *s = 0;

    for (it = args; it != NULL; it = it->next) {
        char * t = my_shell_quote(it->value);
        strcat(s, t);
        if (it->next)
            strcat(s, " ");
        g_free(t);
    }

    return s;
}
#endif

static void guest_exec_child_watch(GPid pid, gint status, gpointer data)
{
    GuestExecInfo *gei = (GuestExecInfo *)data;

    g_debug("guest_exec_child_watch called, pid: %d, status: %u",
            (int32_t)gpid_to_int64(pid), (uint32_t)status);

    gei->status = status;
    qatomic_mb_set(&gei->finished, true);

    g_spawn_close_pid(pid);

    if (gei->with_pty) {
        close(gei->slave_pty);
        gei->slave_pty = -1;
    }

    if (gei->in.cid) {
        aux_remove_channel(gei->in.cid);
    } else {
        if (g_atomic_pointer_get(&gei->in.ch) != NULL) {
            g_io_channel_unref(gei->in.ch);
            g_atomic_pointer_set(&gei->in.ch, NULL);
        }
    }
}

/*
 * Routine called in child process on UNIX/POSIX platforms.
 * Reset ignored signals back to default.
 * Setup master/slave pty pair and redirect stdio for
 * exec with pty (shell enter) command.
 */
static void guest_exec_task_setup(gpointer data)
{
#if !defined(G_OS_WIN32)
    GuestExecInfo *gei = data;
    struct sigaction sigact;

    memset(&sigact, 0, sizeof(struct sigaction));
    sigact.sa_handler = SIG_DFL;

    if (sigaction(SIGPIPE, &sigact, NULL) != 0 ||
        sigaction(SIGTERM, &sigact, NULL) != 0 ||
        sigaction(SIGINT, &sigact, NULL) != 0 )
    {
        slog("sigaction() failed to reset child to dfl: %s", strerror(errno));
        exit(errno);
    }

    if (gei->with_pty) {
        int fd;

        if ((fd = open("/dev/tty", O_RDWR|O_NOCTTY)) >= 0) {
            ioctl(fd, TIOCNOTTY, NULL);
            close(fd);
        }

        if (setsid() < 0) {
            slog("setsid failed: %s", strerror(errno));
            exit(errno);
        }

        if (ioctl(gei->slave_pty, TIOCSCTTY, NULL) < 0) {
            slog("ioctl(TIOCSCTTY) failed: %s", strerror(errno));
            exit(errno);
        }

        if (dup2(gei->slave_pty, STDIN_FILENO) != STDIN_FILENO) {
            slog("dup2 failed for fd=%d: %s", STDIN_FILENO, strerror(errno));
            exit(errno);
        }

        if (dup2(gei->slave_pty, STDOUT_FILENO) != STDOUT_FILENO) {
            slog("dup2 failed for fd=%d: %s", STDOUT_FILENO, strerror(errno));
            exit(errno);
        }

        if (dup2(gei->slave_pty, STDERR_FILENO) != STDERR_FILENO) {
            slog("dup2 failed for fd=%d: %s", STDERR_FILENO, strerror(errno));
            exit(errno);
        }

        close(gei->master_pty);
        close(gei->slave_pty);
    }
#endif
}

static gboolean guest_exec_input_watch(GIOChannel *ch,
        GIOCondition cond, gpointer p_)
{
    GuestExecIOData *p = (GuestExecIOData *)p_;
    gsize bytes_written = 0;
    GIOStatus status;
    GError *gerr = NULL;

    /* nothing left to write */
    if (p->size == p->length) {
        goto done;
    }

    status = g_io_channel_write_chars(ch, (gchar *)p->data + p->length,
            p->size - p->length, &bytes_written, &gerr);

    /* can be not 0 even if not G_IO_STATUS_NORMAL */
    if (bytes_written != 0) {
        p->length += bytes_written;
    }

    /* continue write, our callback will be called again */
    if (status == G_IO_STATUS_NORMAL || status == G_IO_STATUS_AGAIN) {
        return true;
    }

    if (gerr) {
        g_warning("qga: i/o error writing to input_data channel: %s",
                gerr->message);
        g_error_free(gerr);
    }

done:
    g_io_channel_unref(ch);
    g_atomic_pointer_set(&p->ch, NULL);
    g_free(p->data);

    return false;
}

static gboolean guest_exec_output_watch(GIOChannel *ch,
        GIOCondition cond, gpointer p_)
{
    GuestExecIOData *p = (GuestExecIOData *)p_;
    gsize bytes_read;
    GIOStatus gstatus;

    if (cond == G_IO_HUP || cond == G_IO_ERR) {
        goto close;
    }

    if (p->size == p->length) {
        gpointer t = NULL;
        if (!p->truncated && p->size < GUEST_EXEC_MAX_OUTPUT) {
            t = g_try_realloc(p->data, p->size + GUEST_EXEC_IO_SIZE);
        }
        if (t == NULL) {
            /* ignore truncated output */
            gchar buf[GUEST_EXEC_IO_SIZE];

            p->truncated = true;
            gstatus = g_io_channel_read_chars(ch, buf, sizeof(buf),
                                              &bytes_read, NULL);
            if (gstatus == G_IO_STATUS_EOF || gstatus == G_IO_STATUS_ERROR) {
                goto close;
            }

            return true;
        }
        p->size += GUEST_EXEC_IO_SIZE;
        p->data = t;
    }

    /* Calling read API once.
     * On next available data our callback will be called again */
    gstatus = g_io_channel_read_chars(ch, (gchar *)p->data + p->length,
            p->size - p->length, &bytes_read, NULL);
    if (gstatus == G_IO_STATUS_EOF || gstatus == G_IO_STATUS_ERROR) {
        goto close;
    }

    p->length += bytes_read;

    return true;

close:
    g_io_channel_unref(ch);
    g_atomic_pointer_set(&p->ch, NULL);
    return false;
}

#ifndef G_OS_WIN32
/*
 * Find out which shell program does exist in the UNIX/POSIX guest.
 * We return pointer to a local static variable here, this is safe.
 * Return NULL if none of the shells found.
 */
static const char *guest_locate_shell(void)
{
    static const char *const shells[] = {SHELL_CMD, "csh", "sh"};
    int i;

    for (i = 0; i < ARRAY_SIZE(shells); ++i) {
        gchar *abs_shell;
        if ((abs_shell = g_find_program_in_path(shells[i]))) {
            g_free(abs_shell);
            return shells[i];
        }
    }

    return NULL;
}
#endif

GuestExec *qmp_guest_exec(const char *path,
                       bool has_arg, strList *arg,
                       bool has_env, strList *env,
                       bool has_input_data, const char *input_data,
                       bool has_capture_output, bool capture_output,
                       bool has_execute_in_shell, bool execute_in_shell,
                       bool has_with_pty, bool with_pty,
                       bool has_cid_in, int64_t cid_in,
                       bool has_cid_out, int64_t cid_out,
                       bool has_cid_err, int64_t cid_err,
                       bool has_winsize, GuestWinSize *gws,
                       Error **errp)
{
    GPid pid;
    GuestExec *ge = NULL;
    GuestExecInfo *gei;
    char **argv, **envp;
    strList arg_list[1] = {
        { arg, (char*) path },
    };
    strList arg_shell[3] = {
        { &arg_shell[1], (char*) SHELL_CMD },
        /* Windows gspawn() do its own quoting, so we cannot pass
         * cmd /c args to spawning process with properly constructed quotes.
         * As a workaround, we pass args as separate list and rely on
         * gspawn() quoting. It works well for most cases except when
         * program path contains spaces. */
        { &arg_shell[2], (char*) SHELL_ARG },
#ifdef G_OS_WIN32
        { arg, (char*) path },
#else
        { NULL, concat_args(arg_list) }
#endif
    };
    gboolean ret;
    GError *gerr = NULL;
    gint in_fd, out_fd, err_fd;
    GSpawnFlags flags;
    bool has_input = (has_input_data || has_cid_in);
    bool has_output = (has_capture_output && capture_output) || has_cid_out;
    bool has_errput = (has_capture_output && capture_output) || has_cid_err;
    bool run_in_shell = (has_execute_in_shell && execute_in_shell);
    g_autofree uint8_t *input = NULL;
    size_t ninput = 0;

    if (has_input_data) {
        input = qbase64_decode(input_data, -1, &ninput, errp);
        if (!input) {
            return NULL;
        }
    }

    with_pty = (has_with_pty && with_pty);

    if (!strcmp(path, "enter")) {
        run_in_shell = true;
#ifndef G_OS_WIN32
        with_pty = true;
        arg_shell[1].value = (char*)"-i";
        arg_shell[1].next = NULL;
#else
        arg_shell[0].next = NULL;
#endif
    }

#ifndef G_OS_WIN32
    /*
     * If we fail to find a shell, we still put "bash" in here since we
     * want it to be present in the error message.
     */
    if (run_in_shell) {
        char *shell = (char *) guest_locate_shell();
        arg_shell[0].value = shell ? shell : (char *) SHELL_CMD;
    }
#endif

    argv = guest_exec_get_args(run_in_shell ? arg_shell : arg_list, true);
    envp = has_env ? guest_exec_get_args(env, false) : NULL;

    flags = G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD |
        G_SPAWN_SEARCH_PATH_FROM_ENVP;
    if (!has_output) {
        flags |= G_SPAWN_STDOUT_TO_DEV_NULL;
    }
    if (!has_errput) {
        flags |= G_SPAWN_STDERR_TO_DEV_NULL;
    }

    gei = g_new0(GuestExecInfo, 1);
    gei->with_pty = with_pty;
    gei->has_output = has_output;
    gei->has_errput = has_errput;

#ifndef G_OS_WIN32
    if (gei->with_pty) {
        struct winsize ws_data;
        struct winsize *ws = NULL;
        if (has_winsize) {
            ws = &ws_data;
            *ws = (struct winsize) {
                .ws_row = (unsigned short)gws->lines,
                .ws_col = (unsigned short) gws->columns,
            };
        }
        flags |= G_SPAWN_CHILD_INHERITS_STDIN;
        if (openpty(&gei->master_pty, &gei->slave_pty, NULL, NULL, ws) < 0) {
            error_setg(errp, QERR_QGA_COMMAND_FAILED, strerror(errno));
            g_free(gei);
            goto done;
        }
    }
#endif

    ret = g_spawn_async_with_pipes(NULL, argv, envp, flags,
            guest_exec_task_setup, gei, &pid,
            has_input  && !gei->with_pty ? &in_fd  : NULL,
            has_output && !gei->with_pty ? &out_fd : NULL,
            has_errput && !gei->with_pty ? &err_fd : NULL,
            &gerr);
    if (!ret) {
#ifndef G_OS_WIN32
        if (gei->with_pty) {
            close(gei->master_pty);
            close(gei->slave_pty);
        }
#endif
        error_setg(errp, QERR_QGA_COMMAND_FAILED, gerr->message);
        g_error_free(gerr);
        g_free(gei);
        goto done;
    }
#ifndef G_OS_WIN32
    if (gei->with_pty) {
        in_fd = gei->master_pty;
        if (has_output) {
            out_fd = dup(gei->master_pty);
        }
        if (has_errput) {
            err_fd = open("/dev/null", O_RDONLY);
        }
    }
#endif

    ge = g_new0(GuestExec, 1);
    ge->pid = gpid_to_int64(pid);
    guest_exec_info_add(gei, pid);

    g_child_watch_add(pid, guest_exec_child_watch, gei);

    if (has_input) {
        gei->in.data = g_steal_pointer(&input);
        gei->in.data = input;
        gei->in.size = ninput;
#ifdef G_OS_WIN32
        gei->in.ch = g_io_channel_win32_new_fd(in_fd);
#else
        gei->in.ch = g_io_channel_unix_new(in_fd);
#endif
        g_io_channel_set_encoding(gei->in.ch, NULL, NULL);
        g_io_channel_set_buffered(gei->in.ch, false);
        g_io_channel_set_flags(gei->in.ch, G_IO_FLAG_NONBLOCK, NULL);
        g_io_channel_set_close_on_unref(gei->in.ch, true);
        if (has_cid_in) {
            gei->in.cid = cid_in;
            aux_add_channel(0, cid_in, &gei->in.ch);
        } else {
            g_io_add_watch(gei->in.ch, G_IO_OUT, guest_exec_input_watch, &gei->in);
        }
    }

    if (has_output) {
#ifdef G_OS_WIN32
        gei->out.ch = g_io_channel_win32_new_fd(out_fd);
#else
        gei->out.ch = g_io_channel_unix_new(out_fd);
#endif
        g_io_channel_set_encoding(gei->out.ch, NULL, NULL);
        g_io_channel_set_buffered(gei->out.ch, false);
        g_io_channel_set_close_on_unref(gei->out.ch, true);
        if (has_cid_out) {
            gei->out.cid = cid_out;
            aux_add_channel(1, cid_out, &gei->out.ch);
        } else {
            g_io_add_watch(gei->out.ch, G_IO_IN | G_IO_HUP,
                    guest_exec_output_watch, &gei->out);
        }
    }
    if (has_errput) {
#ifdef G_OS_WIN32
        gei->err.ch = g_io_channel_win32_new_fd(err_fd);
#else
        gei->err.ch = g_io_channel_unix_new(err_fd);
#endif
        g_io_channel_set_encoding(gei->err.ch, NULL, NULL);
        g_io_channel_set_buffered(gei->err.ch, false);
        g_io_channel_set_close_on_unref(gei->err.ch, true);
        if (has_cid_err) {
            gei->err.cid = cid_err;
            aux_add_channel(2, cid_err, &gei->err.ch);
        } else {
            g_io_add_watch(gei->err.ch, G_IO_IN | G_IO_HUP,
                    guest_exec_output_watch, &gei->err);
        }
    }

done:
#ifndef G_OS_WIN32
    if (run_in_shell)
        g_free(arg_shell[2].value);
#endif
    g_free(argv);
    g_free(envp);

    return ge;
}

/* Convert GuestFileWhence (either a raw integer or an enum value) into
 * the guest's SEEK_ constants.  */
int ga_parse_whence(GuestFileWhence *whence, Error **errp)
{
    /*
     * Exploit the fact that we picked values to match QGA_SEEK_*;
     * however, we have to use a temporary variable since the union
     * members may have different size.
     */
    if (whence->type == QTYPE_QSTRING) {
        int value = whence->u.name;
        whence->type = QTYPE_QNUM;
        whence->u.value = value;
    }
    switch (whence->u.value) {
    case QGA_SEEK_SET:
        return SEEK_SET;
    case QGA_SEEK_CUR:
        return SEEK_CUR;
    case QGA_SEEK_END:
        return SEEK_END;
    }
    error_setg(errp, "invalid whence code %"PRId64, whence->u.value);
    return -1;
}

GuestHostName *qmp_guest_get_host_name(Error **errp)
{
    GuestHostName *result = NULL;
    g_autofree char *hostname = qemu_get_host_name(errp);

    /*
     * We want to avoid using g_get_host_name() because that
     * caches the result and we wouldn't reflect changes in the
     * host name.
     */

    if (!hostname) {
        hostname = g_strdup("localhost");
    }

    result = g_new0(GuestHostName, 1);
    result->host_name = g_steal_pointer(&hostname);
    return result;
}

GuestTimezone *qmp_guest_get_timezone(Error **errp)
{
    GuestTimezone *info = NULL;
    GTimeZone *tz = NULL;
    gint64 now = 0;
    gint32 intv = 0;
    gchar const *name = NULL;

    info = g_new0(GuestTimezone, 1);
    tz = g_time_zone_new_local();
    if (tz == NULL) {
        error_setg(errp, QERR_QGA_COMMAND_FAILED,
                   "Couldn't retrieve local timezone");
        goto error;
    }

    now = g_get_real_time() / G_USEC_PER_SEC;
    intv = g_time_zone_find_interval(tz, G_TIME_TYPE_UNIVERSAL, now);
    info->offset = g_time_zone_get_offset(tz, intv);
    name = g_time_zone_get_abbreviation(tz, intv);
    if (name != NULL) {
        info->has_zone = true;
        info->zone = g_strdup(name);
    }
    g_time_zone_unref(tz);

    return info;

error:
    g_free(info);
    return NULL;
}

GuestFileRead *qmp_guest_file_read(int64_t handle, bool has_count,
                                   int64_t count, Error **errp)
{
    GuestFileHandle *gfh = guest_file_handle_find(handle, errp);
    GuestFileRead *read_data;

    if (!gfh) {
        return NULL;
    }
    if (!has_count) {
        count = QGA_READ_COUNT_DEFAULT;
    } else if (count < 0 || count > GUEST_FILE_READ_COUNT_MAX) {
        error_setg(errp, "value '%" PRId64 "' is invalid for argument count",
                   count);
        return NULL;
    }

    read_data = guest_file_read_unsafe(gfh, count, errp);
    if (!read_data) {
        slog("guest-file-write failed, handle: %" PRId64, handle);
    }

    return read_data;
}

#ifdef G_OS_WIN32
#include <tlhelp32.h>

static int kill_process_tree(int64_t pid)
{
    int rc = 0;
    PROCESSENTRY32 pe;
    GList *l = NULL;

    HANDLE snap = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
    if (snap == INVALID_HANDLE_VALUE) {
        return GetLastError();
    }

    l = g_list_append(l, GUINT_TO_POINTER(pid));

    pe.dwSize = sizeof(PROCESSENTRY32);
    bool added, o;
    do {
        added = false;
        for (o = Process32First(snap, &pe); o; o = Process32Next(snap, &pe)) {
            if (g_list_find(l, GUINT_TO_POINTER(pe.th32ParentProcessID)) &&
                    !g_list_find(l, GUINT_TO_POINTER(pe.th32ProcessID))) {
                l = g_list_append(l, GUINT_TO_POINTER(pe.th32ProcessID));
                added = true;
            }
        }
    } while (added);

    for (o = Process32First(snap, &pe); o; o = Process32Next(snap, &pe)) {
        if (g_list_find(l, GUINT_TO_POINTER(pe.th32ProcessID))) {
            g_debug("killing pid=%u ppid=%u name=%s",
                (guint)pe.th32ProcessID, (guint)pe.th32ParentProcessID,
                pe.szExeFile);
        }
    }

    CloseHandle(snap);

    GList *e;
    for (e = l; e; e = e->next) {
        pid = GPOINTER_TO_UINT(e->data);
        HANDLE h = OpenProcess(PROCESS_TERMINATE, FALSE, pid);
        if (h == INVALID_HANDLE_VALUE) {
            if (!rc) {
                rc = GetLastError();
                if (rc == ERROR_FILE_NOT_FOUND) {
                    rc = 0;
                }
            }
            continue;
        }
        TerminateProcess(h, 255);
        CloseHandle(h);
    }

    g_list_free(l);

    return rc;
}
#endif

void qmp_guest_exec_terminate(int64_t pid, Error **err)
{
    GuestExecInfo *gei;

    slog("guest-exec-terminate called, pid: %u", (uint32_t)pid);

    gei = guest_exec_info_find(pid, err);
    if (gei == NULL) {
        slog("guest-exec-terminate, pid: %u not found", (uint32_t)pid);
        return;
    }

    if (qatomic_mb_read(&gei->finished)) {
        return;
    }

#ifdef G_OS_WIN32
    int rc = kill_process_tree(pid);
    if (rc != 0) {
        char buf[64];
        sprintf(buf, "win32 err %d", rc);
        error_setg(err, QERR_QGA_COMMAND_FAILED, buf);
    }
#else
    if (kill(pid, SIGKILL) < 0) {
        if (errno != ESRCH) {
            error_setg(err, QERR_QGA_COMMAND_FAILED, strerror(errno));
        }
    }
#endif
}

void qmp_guest_exec_set_parameters(int64_t pid,
    bool has_winsize, GuestWinSize *gws,
    Error **err)
{
    GuestExecInfo *gei;

    slog("guest-exec-set-parameters called, pid: %u", (uint32_t)pid);

    gei = guest_exec_info_find(pid, err);
    if (gei == NULL) {
        slog("guest-exec-set-parameters, pid: %u not found", (uint32_t)pid);
        return;
    }

#ifndef G_OS_WIN32
    if (gei->with_pty && has_winsize) {
        struct winsize ws = {
            (unsigned short) gws->lines,
            (unsigned short) gws->columns
        };
        ioctl(gei->slave_pty, TIOCSWINSZ, &ws);
    }
#endif
}
